#! /usr/bin/env python3

# hit percentage > 50%, evalue < 0.01
# index 15, 14, 5, and 6
# total genes for taxon of interest, total each pfam for taxon of interest
    # acquire

# NEED to make enrichment analysis consider all Pfam hits for a gene if it isn't already
# NEED top hit argument
# NEED specific GO term argument

from scipy.stats import hypergeom
from mycotools.lib.biotools import fa2dict, gff3Comps, gff2list
from mycotools.lib.kontools import format_path, eprint, mkOutput
from mycotools.lib.dbtools import mtdb
from collections import Counter, defaultdict
from tqdm import tqdm
import multiprocessing as mp, os, argparse, sys, re

def collectGffs(omes, top_dir):
    clusGFFs = {
        ome: top_dir + ome + '/' + ome + '.gff3' for ome in omes \
        if os.path.isfile(top_dir + ome + '/' + ome + '.gff3')
        }
    return tuple(clusGFFs.items())

def parseGff(ome, gff_path, aliasComp = re.compile(gff3Comps()['Alias'])):
    genes = {}
    gff = gff2list(gff_path)
    for entry in gff:
        clusID = re.search(r'clusID=([^;]+)', entry['attributes'])[1]
        alias = aliasComp.search(entry['attributes'])[1]
        genes[alias] = clusID
    return ome, tuple(genes.items())

def extractGO( file_, goTerm = 'GO:0008152' ):
    '''From the GO file, compile all GO terms. For each GO, add to `godict` as the key and the set of GO terms it is as value.
    For each GO term of each key in `godict`, check all of their sub GO terms to see if `goTerm` is within one. If so, append
    it to the output.'''

    goSet = set()
    goSet.add(goTerm)
    count = 1

    godict = {}
    goterm = re.compile(r'GO\:\d+')
    with open( file_, 'r' ) as raw:
        for line in raw:
            if line.startswith('[Term]'):
                goid = None
            elif line.startswith('id:'):
                goid = goterm.search( line )
                if goid:
                    goid = goid[0]
                    godict[goid] = set()
            elif line.startswith('is_a:'):
                goisa = goterm.search( line )
                if goisa:
                    goisa = goisa[0]
                    godict[goid].add( goisa )

    todel = []
    for goid in godict:
        for sub_goid in godict[goid]:
            temp_goSet = set(sub_goid)
            temp_goList = [sub_goid]
            count = 1
            while count == 1:
                del_list = []
                for index in range(len(temp_goList)):
                    del_list.append(index)
                    id1 = str(temp_goList[index])
                    temp_goList.extend(godict[id1])
                    temp_goSet = temp_goSet | set(godict[id1])
                    if goTerm in temp_goSet:
                        goSet = goSet | {goid, sub_goid}
                        count = 0
                        break
                del_list.sort(reverse = True)
                for del_term in del_list:
                    del temp_goList[del_term]
                if len(temp_goList) == 0:
                    count = 0

    return goSet


def parsePfam2GO(pfam2go_path, pfam2def, go_sets):
    pfam2go, go2onto = defaultdict(list), {}
    go2pfam = defaultdict(list)
    with open(pfam2go_path, 'r') as raw:
        for line in raw:
            try:
                pfam = re.search(r'PF\d+', line)[0]
                go = re.search(r'GO\:\d+', line)[0]
                onto = re.search(r'GO\:([^;]+)', line)[1]
                try:
                     definition = pfam2def[pfam]
                except KeyError: # not used, who cares
                    continue
                for go_term, go_set in go_sets.items():
                    if go in go_set:
                        go2pfam[go_term].append(definition)
                pfam2go[definition].append(go)
                go2onto[go] = onto
            except TypeError:
                continue
    return pfam2go, go2onto, dict(go2pfam)
            

def countGenes(ome, proteome_path):
    return ome, len(fa2dict(proteome_path))

def clusterGenes(ome, info_path):
    genes = {}
    try:
        with open(info_path, 'r') as raw:
            for line in raw:
                if line.startswith('#'):
                    continue
                d = line.split('\t')
                clus, genesT = d[0], d[2].split(',')
                genes = {**genes, **{x: clus for x in genesT}}
    except FileNotFoundError:
        pass
    return ome, tuple(genes.items())

def parsePfam(ome, pfam_res, clusGenes, hlg_genes, top_hit = True):
    gene2pfam = defaultdict(dict)
    clusGenes = dict(clusGenes)
    hlgGenes = dict(hlg_genes)
    pfam2def = {}
    try:
        with open(pfam_res, 'r') as raw:
            for line in raw:
                if line.startswith('#'):
                    continue
                data = line.rstrip().split()
                g, p, q, b = data[0], data[4], data[3], float(data[7])
                
                # this is where a post-hoc threshold would be applied
                if q not in gene2pfam[g]: # in case there is a duplicate
                # for the same gene, add in the better score 
                    gene2pfam[g][q] = b
                else:
                    if b > gene2pfam[g][q]:
                        gene2pfam[g][q] = b
                pfam2def[p[:p.find('.')]] = q
    except FileNotFoundError:
        return ome, None, None, None, None
    
    if top_hit:
        gene2pfam = {k: dict(sorted(v.items(), key = lambda x: x[1], reverse = True)) \
                     for k, v in gene2pfam.items()}
        gene2pfam = {k: {list(v.keys())[0]: v[list(v.keys())[0]]} for k, v in gene2pfam.items()}
    gene2pfam, clusGenes2pfam = dict(gene2pfam), {}
    for gene, clus in clusGenes.items():
        try:
            clusGenes2pfam[gene] = tuple([gene2pfam[gene], clus])
        except KeyError: # no pfam acc found, should be rarer
            pass
    hlg_genes2pfam = {}
    for gene, clus in hlgGenes.items():
        try:
            hlg_genes2pfam[gene] = tuple([gene2pfam[gene], clus])
        except KeyError:
            pass
    
    return ome, tuple(gene2pfam.items()), \
        tuple(clusGenes2pfam.items()), tuple(hlg_genes2pfam.items()), tuple(pfam2def.items())
#        tuple([tuple([gene, pfams.items()]) for gene, pfams in clusGenes2pfam.items()])


def calcP(pfam, totCount, clusCount, totalGenes, totalClusGenes, pfamInClus):
    pval = hypergeom.sf(
        clusCount-1, totalGenes, totCount, totalClusGenes
        )
    return pfam, pval, totCount/totalGenes, clusCount/totalClusGenes


def GetPopResultsGO(
    omes2genes2pfams, omes2hlgs2pfam, omes2clus2pfam, totalGenes, total_hlg_genes, totalClusGenes, 
    pfam2go, go2pfam, pool
    ):
    totalPfams, clusPfams, pfam2clusters = Counter(), Counter(), defaultdict(list)
    hlg_pfams, pfam2hlgs = Counter(), defaultdict(list)
    totalGOs, clusGOs, go2clusters = Counter(), Counter(), defaultdict(list)
    hlg_gos, go2hlgs = Counter(), defaultdict(list)
    total_ref_gos, ref_go2clusters = Counter(), defaultdict(list)
    ref_go2hlgs = defaultdict(list)
    sp_gos = Counter()
    for ome, genes2pfam in tqdm(omes2genes2pfams.items(), total = len(omes2genes2pfams)):
        clusGenes2pfam = omes2clus2pfam[ome]
        hlgGenes2pfam = omes2hlgs2pfam[ome]
        tPfams, tClusPfams, t_hlg_pfams = [], [], []
        tGOs, tClusGOs, t_hlg_gos = [], [], []
        t_s_gos, t_s_clus_gos, t_s_hlg_gos = defaultdict(int), defaultdict(int), defaultdict(int)
        # count all instances
        for pfams in list(genes2pfam.values()):
            tPfams.extend(pfams)
            for pfam in pfams:
                go_terms = pfam2go[pfam]
                if not go2pfam:
                    tGOs.extend(go_terms)
                else:
                    for ref_go, pfam_set in go2pfam.items():
                        if pfam in pfam_set:
                            t_s_gos[ref_go] += 1
             
        for pfams, hlg in list(hlgGenes2pfam.values()):
            for pfam in pfams:
                pfam2hlgs[pfam].append(f'{ome}_{hlg}')
                if not go2pfam:
                    for go in pfam2go[pfam]:
                        go2hlgs[go].append(f'{ome}_{hlg}')
                        t_hlg_gos.append(go)
                else:
                    for ref_go, pfam_set in go2pfam.items():
                        if pfam in pfam_set:
                            t_s_hlg_gos[ref_go] += 1
                            ref_go2hlgs[ref_go].append(f'{ome}_{hlg}')
                t_hlg_pfams.append(pfam)   

        for pfams, clus in list(clusGenes2pfam.values()):
#            if any('F-box' in x for x in pfams) and any('WD40' in x for x in pfams):
 #               print(clus, flush = True)
            for pfam in pfams:
                pfam2clusters[pfam].append(ome + '_' + clus)
                if not go2pfam:
                    for go in pfam2go[pfam]:
                        go2clusters[go].append(ome + '_' + clus)
                        tClusGOs.append(go)
                else:
                    for ref_go, pfam_set in go2pfam.items():
                        if pfam in pfam_set:
                            t_s_clus_gos[ref_go] += 1
                            ref_go2clusters[ref_go].append(ome + '_' + clus)
                tClusPfams.append(pfam)
        totalPfams += Counter(tPfams)
        clusPfams += Counter(tClusPfams)
        hlg_pfams += Counter(t_hlg_pfams)
        totalGOs += Counter(tGOs)
        clusGOs += Counter(tClusGOs)
        hlg_gos += Counter(t_hlg_gos)
        total_ref_gos += Counter(t_s_gos)

    cvt_c, cvh_c, hvt_c = [], [], []
#    p_cvt_pval, p_cvh_pval = [], []
    for pfam, clusCount in clusPfams.items():
        totCount = totalPfams[pfam]
        hlgCount = hlg_pfams[pfam]
        pfamInClus = len(set(pfam2clusters[pfam]))
 #       p_cvt_pval.append(calcP(pfam, totCount, clusCount, totalGenes, totalClusGenes, pfamInClus))
  #      p_cvh_pval.append(calcP(pfam, hlgCount, clusCount, total_hlg_genes, totalClusGenes, pfamInClus))
        cvt_c.append([pfam, totCount, clusCount, totalGenes, totalClusGenes, pfamInClus])
        cvh_c.append([pfam, hlgCount, clusCount, total_hlg_genes, totalClusGenes, pfamInClus])
    p_cvt_pval = pool.starmap(calcP, cvt_c)
    p_cvh_pval = pool.starmap(calcP, cvh_c)

#    p_hvt_pval = []    
    for pfam, hlgCount in hlg_pfams.items():
        totCount = totalPfams[pfam]
        hlgCount = hlg_pfams[pfam]
        pfam_in_hlg = len(set(pfam2hlgs[pfam]))
#        p_hvt_pval.append(calcP(pfam, totCount, hlgCount, totalGenes, total_hlg_genes, pfam_in_hlg))
        hvt_c.append([pfam, totCount, hlgCount, totalGenes, total_hlg_genes, pfam_in_hlg])
    p_hvt_pval = pool.starmap(calcP, hvt_c)

    cvt_c, cvh_c, hvt_c = [], [], []
    if not go2pfam:
        for go, clusCount in clusGOs.items():
            totCount = totalGOs[go]
            hlgCount = hlg_gos[go]
            goInClus = len(set(go2clusters[go]))
            cvt_c.append([go, totCount, clusCount, totalGenes, totalClusGenes, goInClus])
            cvh_c.append([go, hlgCount, clusCount, total_hlg_genes, totalClusGenes, goInClus])
        g_cvt_pval = pool.starmap(calcP, cvt_c)
        g_cvh_pval = pool.starmap(calcP, cvh_c)
        for go, hlgCount in hlg_gos.items():
            totCount = totalGOs[go]
            hlgCount = hlg_gos[go]
            go_in_hlg = len(set(go2hlgs[go]))
            hvt_c.append([go, totCount, hlgCount, totalGenes, total_hlg_genes, go_in_hlg])
        g_hvt_pval = pool.starmap(calcP, hvt_c)
        r_cvt_pval = g_cvt_pval
        r_cvh_pval = g_cvh_pval
        r_hvt_pval = g_hvt_pval
    else: 
 #       r_cvt_pval, r_cvh_pval, r_hvt_pval = [], [], []
        for ref_go, gos in ref_go2clusters.items():
            totCount = total_ref_gos[ref_go]
            ref_in_clus = len(set(gos))
            ref_in_hlg = len(set(ref_go2hlgs[ref_go]))
 #           r_cvt_pval.append(calcP(ref_go, totCount, ref_in_clus, totalGenes, totalClusGenes, ref_in_clus))
#            r_cvh_pval.append(calcP(ref_go, ref_in_hlg, ref_in_clus, total_hlg_genes, totalClusGenes, ref_in_clus))
            cvt_c.append([ref_go, totCount, ref_in_clus, totalGenes, totalClusGenes, ref_in_clus])
            cvh_c.append([ref_go, ref_in_hlg, ref_in_clus, total_hlg_genes, totalClusGenes, ref_in_clus])
        r_cvt_pval = pool.starmap(calcP, cvt_c)
        r_cvh_pval = pool.starmap(calcP, cvh_c)
    
        for ref_go, gos in ref_go2hlgs.items():
            totCount = total_ref_gos[ref_go]
            ref_in_hlg = len(set(gos))
#            r_hvt_pval.append(calcP(ref_go, totCount, ref_in_hlg, totalGenes, total_hlg_genes, ref_in_hlg))
            hvt_c.append([ref_go, totCount, ref_in_hlg, totalGenes, total_hlg_genes, ref_in_hlg])
        r_hvt_pval = pool.starmap(calcP, hvt_c)
    

    return [p_cvt_pval, p_cvh_pval, p_hvt_pval], \
           [None, None, None], \
           [r_cvt_pval, r_cvh_pval, r_hvt_pval]
#           [g_cvt_pval, g_cvh_pval, g_hvt_pval], \



# need total genes, per pfam counts, pfams in clusters counts, genes 
def GetPopResults(omes2genes2pfams, omes2clus2pfam, totalGenes, totalClusGenes, pool):
    totalPfams, clusPfams, pfam2clusters = Counter(), Counter(), defaultdict(list)
    for ome, genes2pfam in omes2genes2pfams.items():
        clusGenes2pfam = omes2clus2pfam[ome]
        tPfams, tClusPfams = [], []
        for pfams in list(genes2pfam.values()):
            tPfams.extend(pfams)
        for pfams, clus in list(clusGenes2pfam.values()):
            for pfam in pfams:
#                print(pfam, flush = True)
                if pfam == 'DUF3435' or pfam.lower() == 'plavaka':
                    print(ome, clus, pfam, flush = True)
                pfam2clusters[pfam].append(ome + '_' + clus)
                tClusPfams.append(pfam)
        totalPfams += Counter(tPfams)
        clusPfams += Counter(tClusPfams)

    cmds = []
    for pfam, clusCount in clusPfams.items():
        totCount = totalPfams[pfam]
        pfamInClus = len(set(pfam2clusters[pfam]))
        cmds.append([pfam, totCount, clusCount, totalGenes, totalClusGenes, pfamInClus])

    pvalRes = pool.starmap(calcP, cmds)

    return pvalRes

def WriteOutput(scoresDict, thresh, out_file, go2onto = None, go2pfam = False, sort = False):
# out_dir + taxon.lower() + '.scores.tsv'
    try:
        alpha = thresh/len(scoresDict) # bonferroni correction
    except ZeroDivisionError:
        eprint('\t\tWARNING: no hits', flush = True)
        return None

    sig, insig = [], []
    with open(out_file + '.tmp', 'w') as out:
        if go2onto:
            out.write(
                '#go\ttotal\tclusters\tp\tsignificance (alpha =' + \
                str(alpha) + ')\tontology\n'
                )
        elif go2pfam:
            out.write(f'#go\tcompared\tcomparee\tp\tsignificance (alpha ={alpha})\n')
        else:
            out.write('#pfam\tcompared\tcomparee\tp\tsignificance (alpha=' + str(alpha) + ')\n')
        for q, val, tot, clusTot in scoresDict:
            if val <= alpha: # significant
                out.write(f'{q}\t{tot}\t{clusTot}\t{val}\t1')
#                sig.append([q, tot, clusTot, val, 1])
            else: # not significant
                out.write(f'{q}\t{tot}\t{clusTot}\t{val}\t0')
            if go2onto:
                out.write(f'\t{go2onto[q]}\n')
            else:
                out.write('\n')

    os.rename(out_file + '.tmp', out_file)

    if sort:
        sig = []
        insig = []
        with open(out_file, 'r') as raw:
            for line in raw:
                d = line.rstrip().split()
                if d[4] == '1':
                    sig.append(d)
                else:
                    insig.append(d)
        with open(out_file + '.tmp', 'w') as out: 
            sig = sorted(sig, key = lambda x: float(x[2]), reverse = True)
            out.write('\n'.join(['\t'.join([str(x) for x in y]) for y in sig]) + '\n')
            out.write('\n'.join(['\t'.join([str(x) for x in y]) for y in insig]))
        os.rename(out_file + '.tmp', out_file)


def main(
    db, thresh, rank, ome_dir, pfam_dir, out_dir,
    pfam2go_path, go_file, go_terms = [], top_hit = True, 
    pool = mp.Pool(processes = 1), minimum = 0
    ):

    db = db.set_index()
    ranks = defaultdict(list)
    for ome, row in db.items():
        taxon = row['taxonomy'][rank]
        ranks[taxon].append(ome)

    ranks = {k: v for k, v in ranks.items() if len(v) > minimum}

    print('\nChecking input', flush = True)
    for taxon, omes in ranks.items():
        todel = []
        for i, ome in enumerate(omes):
 #           if ome_dir:
            if not os.path.isfile(ome_dir + ome + '/hlg.tsv'):
                todel.append(i)
            elif not os.path.isfile(ome_dir + ome + '/gcf.tsv'):
                eprint(f'\tWARNING: {ome} gcf.tsv missing, assuming no GCFs', flush = True)
#                sys.exit(14)
#            elif not os.path.isfile(as_dir + ome + '/' + ome + '.gff3'):
 #               todel.append(i)
        for i in reversed(todel):
            del omes[i]

    go_sets = {}
    for go_term in go_terms:
        go_sets[go_term] = extractGO(go_file, go_term)

    ranks = {k: v for k, v in sorted(ranks.items(), key = lambda x: len(x[1]))}
    for taxon, omes in ranks.items():
        if not omes:
            continue
        print('\n' + taxon + ': ' + str(len(omes)) + ' omes', flush = True)
        print('\tGenes', flush = True)
        taxonGenes, taxon_hlg_genes, taxonClusGenes = 0, 0, 0
        countres = pool.starmap(
            countGenes, 
            [[ome, db[ome]['faa']] for ome in omes]
            )
        for ome, tot in countres:
            taxonGenes += tot

      #  if not as_dir:
        print('\tHLGs', flush = True)
        hlgres = pool.starmap(
            clusterGenes,
            [[ome, ome_dir + ome + '/hlg.tsv'] for ome in omes]
            )

        print('\tGCFs', flush = True)
        gcfres = pool.starmap(
            clusterGenes,
            [[ome, f'{ome_dir}{ome}/gcf.tsv'] for ome in omes]
            )
#        elif as_dir:
 #           gffs = collectGffs(omes, as_dir)
  #          clusres = pool.starmap(
   #             parseGff,
    #            [[ome, gff_path] for ome, gff_path in gffs]
     #           )

        print('\tPfams', flush = True)
        ome2clus_genes, ome2hlg_genes = {}, {}
        for ome, genes2clus in hlgres:
            ome2hlg_genes[ome] = genes2clus
            taxon_hlg_genes += len(genes2clus)
        for ome, genes2clus in gcfres:
            ome2clus_genes[ome] = genes2clus
            taxonClusGenes += len(genes2clus)

        pfamres = pool.starmap(
            parsePfam,
            [[ome, f'{pfam_dir}{ome}_pfam.out', ome2clus_genes[ome], ome2hlg_genes[ome], top_hit] \
            for ome in omes]
            )

        ome2genes2pfam, ome2clus_genes2pfam, ome2hlgGenes2pfam, pfam2def = {}, {}, {}, {}
        for ome, tgene2pfam, tclusGenes2pfam, thlgGenes2pfam, tpfam2def in pfamres:
            if not tgene2pfam:
                eprint(f'\t\tWARNING: {ome} missing annotation', flush = True)
                continue
            ome2genes2pfam[ome] = {x[0]: dict(x[1]) for x in tgene2pfam}
            ome2clus_genes2pfam[ome] = {x[0]: x[1] for x in tclusGenes2pfam}
            ome2hlgGenes2pfam[ome] = {x[0]: x[1] for x in thlgGenes2pfam}
            pfam2def = {**pfam2def, **dict(tpfam2def)}

#        if as_dir:
 #           out_base = out_dir + taxon.lower() + '.as'
  #      else:
        out_base = out_dir + taxon.lower()

        print(f'\tEnrichment {out_base}*', flush = True)
        if pfam2go_path:
            pfam2go, go2onto, go2pfam = parsePfam2GO(pfam2go_path, pfam2def, go_sets)
            pfamScores, goScores, ref_scores = GetPopResultsGO(
                ome2genes2pfam, ome2hlgGenes2pfam, ome2clus_genes2pfam, taxonGenes, 
                taxon_hlg_genes, taxonClusGenes, pfam2go, go2pfam, pool
                )
            print(f'\tWriting', flush = True)
            WriteOutput(pfamScores[0], thresh, out_base + '_cvt.pfam.tsv')
            WriteOutput(pfamScores[1], thresh, out_base + '_cvh.pfam.tsv')
            WriteOutput(pfamScores[2], thresh, out_base + '_hvt.pfam.tsv')
#            WriteOutput(goScores[0], thresh, out_base + '_cvt.go.tsv', go2onto)
 #           WriteOutput(goScores[1], thresh, out_base + '_cvh.go.tsv', go2onto)
  #          WriteOutput(goScores[2], thresh, out_base + '_hvt.go.tsv', go2onto)
            if go2pfam:
                WriteOutput(ref_scores[0], thresh, out_base + '_cvt.ref.go.tsv', go2pfam = True)
                WriteOutput(ref_scores[1], thresh, out_base + '_cvh.ref.go.tsv', go2pfam = True)
                WriteOutput(ref_scores[2], thresh, out_base + '_hvt.ref.go.tsv', go2pfam = True)
        else:
            scoresDict = GetPopResults(ome2genes2pfam, ome2clus_genes2pfam, 
                                       taxonGenes, taxonClusGenes, pool)
            print(f'\tWriting', flush = True)

            WriteOutput(scoresDict, thresh, out_base + '.gcf.pfam.tsv', go2onto = None, sort = True)
            scoresDict = GetPopResults(ome2genes2pfam, ome2hlgGenes2pfam, 
                                       taxonGenes, taxonClusGenes, pool)

            WriteOutput(scoresDict, thresh, out_base + '.hlg.pfam.tsv', go2onto = None, sort = True)

def cli():

    parser = argparse.ArgumentParser(description = 'cloci2pfam enrichment')
    parser.add_argument('-a', '--alpha', help = 'Overall Bonferonni corrected alpha; DEFAULT: 0.05',
        type = float, default = 0.05)
    parser.add_argument('-d', '--mtdb', help = 'CLOCI mycotools db input', required = True)
    parser.add_argument('-r', '--rank', help = 'taxon rank', required = True)
    parser.add_argument('-c', '--cloci', help = 'CLOCI output dir', required = True)
#    parser.add_argument('-a', '--antismash_dir', help = 'antiSMASH dir with ome subfolders')
    parser.add_argument('-p', '--pfam_dir', help = 'Pfam tbl output dir', required = True)
    parser.add_argument('-m', '--minimum', help = 'Minimum omes to consider rank', 
                        type = int, default = 0)
    parser.add_argument('-p2g', '--pfam2go', help = 'pfam2go file for GO')
    parser.add_argument('-g', '--go_terms', help = '[-gf] Specific GO terms to mine')
    parser.add_argument('-gf', '--go_obo', help = 'GO ontology file')
    parser.add_argument('-t', '--top_hit', action = 'store_true', 
                        help = 'Only consider top Pfam hit')
    parser.add_argument('-o', '--out_dir')
    parser.add_argument('--cpu', type = int, default = 1)
    args = parser.parse_args()

    taxon_set = {'kingdom', 'phylum', 'subphylum', 'class', 'order', 'family', 'genus', 'species'}
    if args.rank.lower() not in taxon_set:
        raise ValueError('Taxon must be in ' + str(taxon_set))
#    elif (args.cloci and args.antismash_dir) or (not args.cloci and not args.antismash_dir):
 #       raise ValueError('-o2c OR -as must be inputted')

    db = mtdb(format_path(args.mtdb))
    #if args.cloci:
    ome_dir = format_path(args.cloci + 'ome/', force_dir = True)
   #     as_dir = None
  #  else:
 #       ome_dir = None
#        as_dir = format_path(args.antismash_dir)
    pfam_dir = format_path(args.pfam_dir)

    if not args.out_dir:
#        if os.path.basename(os.path.dirname(format_path(args.cloci)[:-1])) == 'ome':
 #           out_dir = mkOutput(format_path('./'), 'cloci2enrichment')
  #      else:
        out_dir = mkOutput(format_path(args.cloci), 'cloci2enrichment', 
                             suffix = None)
    else:
        out_dir = mkOutput(format_path(out_dir), 'cloci2enrichment')

    if args.pfam2go:
        pfam2go_path = format_path(args.pfam2go)
    else:
        pfam2go_path = None

    if args.go_terms:
        go_terms = args.go_terms.replace('"','').replace("'",'').replace(',',' ').split()
        go_file = format_path(args.go_obo)
    else:
        go_terms, go_file = [], None

    pool = mp.Pool(processes = args.cpu)
    main(db, args.alpha, args.rank.lower(), ome_dir, pfam_dir, out_dir, pfam2go_path, 
         go_file, go_terms, pool = pool, top_hit = args.top_hit, minimum = args.minimum)
    pool.close()
    pool.join()

    sys.exit(0)


if __name__ == '__main__':
    cli()



import os
import sys
import pickle
import datetime
import subprocess
import multiprocessing as mp
from tqdm import tqdm
from itertools import chain
from collections import defaultdict
from mycotools.acc2fa import dbmain as acc2fa
from mycotools.lib.biotools import gff2list
from mycotools.lib.kontools import multisub, tardir, collect_files, checkdir, eprint
from cloci.lib.input_parsing import compileCDS2, hg_fa_mngr
from cloci.lib.treecalcs import calc_tmd

# NEED todel_hg to remove fully incompatible GCL HGs from HLG_HGXs
# NEED an option to import OrthoFinder pairwise alignments
	# not recommended, need higher max target sequences than OrthoFinder 
def parse_algn_wpos(hgx_dir, hg, hgx_genes):
    """Parse alignment data that contains percent positives"""
    qlen, hit_len = len(hgx_genes), 0
    with open(f'{hgx_dir}{hg}.out', 'r') as raw:
        # open the blast results
        gene2algn = defaultdict(list)
        for line in raw:
            d = line.rstrip().split('\t')
            # need to adjust to just use d when safe
            q, s, p, i = d[0], d[1], d[2], d[3]
            if q in hgx_genes:
                gene2algn[q].append((s, float(i), float(p)))
                hit_len = len(gene2algn)
            elif hit_len == qlen:
                break
    return gene2algn


def parse_algn_wopos(hgx_dir, hg, hgx_genes):
    """Parse alignment data that does not contain percent positives (MMseqs)"""
    qlen, hit_len = len(hgx_genes), 0
    with open(f'{hgx_dir}{hg}.out', 'r') as raw:
        # open the blast results
        gene2algn = defaultdict(list)
        for line in raw:
            d = line.rstrip().split('\t')
            # need to adjust to just use d when safe
            q, s, i = d[0], d[1], d[2]
            if q in hgx_genes:
                gene2algn[q].append((s, float(i)))
                hit_len = len(gene2algn)
            elif hit_len == qlen:
                break
    return gene2algn


def get_top_hits_wpos(gene, hlg, hgx_genes, ome_set, hits, gene2algn, res, ome):
    """Identify the top hits of a query by sorting based on alignment
    identity"""
    ome_hits = set()
    ome_len = len(ome_set) - 1
    for sbj_gene, sbj_id, sbj_pos in gene2algn[gene]:
        sbj_ome = sbj_gene[:sbj_gene.find('_')]
    # if the subject is also in the checked queries and it is not
    # a recent paralog of the query species
        if sbj_gene in hgx_genes and sbj_ome != ome:
        # grab the hit similarity
            res[hlg][gene][1][sbj_gene] = (sbj_id, sbj_pos,)
            hits.add(sbj_gene)
            ome_hits.add(sbj_ome)
            if ome_len == len(ome_hits):
                return hits, res 
    # add one to the failed count
        elif sbj_ome not in ome_set: 
            res[hlg][gene][0] += 1 
    
    if len(ome_hits) < ome_len:
        # missing gene and no paralog
        res[hlg][gene][2] = True 

    return hits, res 
        
def get_top_hits_wopos(gene, hlg, hgx_genes, ome_set, hits, gene2algn, res, ome):
    """Identify the top hits of a query by sorting based on alignment
    identity (MMseqs)"""
    ome_len = len(ome_set) - 1
    ome_hits = set()
    for sbj_gene, sbj_id in gene2algn[gene]:
        sbj_ome = sbj_gene[:sbj_gene.find('_')]
        # if the subject is also in the checked queries and it is not
        # a recent paralog of the query species
        if sbj_gene in hgx_genes and sbj_ome != ome:
            # grab the hit similarity
            res[hlg][gene][1][sbj_gene] = sbj_id
            hits.add(sbj_gene)
            ome_hits.add(sbj_ome)
            if len(ome_hits) == ome_len:
                return hits, res
        # add one to the failed count
        elif sbj_ome not in ome_set: 
            res[hlg][gene][0] += 1 
        
    if len(ome_hits) < ome_len:
        # missing gene and no paralog
        # biased against convergent assimilation
        res[hlg][gene][2] = True 

    return hits, res


def sim_calc_wpos(res):
    """Calclate the minimum similarity for identity and positives across the genes
    within a shared HG"""
    omeScores = defaultdict(dict)
    ids_y_pos = defaultdict(dict)
    for hg in res:
        for ome, data in res[hg].items():
            t_gcl, id_y_pos_data = data
            omeScores[ome][hg] = t_gcl
            ids = [x[0] for x in id_y_pos_data.values()]
            pos = [x[1] for x in id_y_pos_data.values()]
            if ids:
                ids_y_pos[hg][ome] = (min(ids), min(pos),)
            else:
                ids_y_pos[hg][ome] = (0, 0,)
    return omeScores, ids_y_pos


def sim_calc_wopos(res):
    """Calclate the minimum similarity for identity across the genes
    within a shared HG"""
    omeScores = defaultdict(dict)
    ids_dict = defaultdict(dict)
    for hg in res:
        for ome, data in res[hg].items():
            t_gcl, id_data = data
            omeScores[ome][hg] = t_gcl
            ids = list(id_data.values())
            if ids:
                ids_dict[hg][ome] = min(ids)
            else:
                ids_dict[hg][ome] = 0
    return omeScores, ids_dict

# deprecated
def mms_calcs_wpos(ids_y_pos):
    # average minimum identity / minimum positive; for each gene homolog group
    hlg_min_id, hlg_min_pos, total = 0, 0, 0
    for ome_dict in ids_y_pos.values():
        hlg_min_id += sum([x[0] for x in ome_dict.values()])
        hlg_min_pos += sum([x[1] for x in ome_dict.values()])
        total += len(ome_dict)
    try:
        hlg_min_id /= total
        hlg_min_pos /= total
    except ZeroDivisionError:
        hlg_min_id, hlg_min_pos = 0, 0
    return hlg_min_id, hlg_min_pos

# deprecated
def mms_calcs_wopos(ids_dict):
    # average minimum identity; for each gene homolog group
    hlg_min_id, total = 0, 0
    for ome_dict in ids_dict.values():
        hlg_min_id += sum(ome_dict.values())
        total += len(ome_dict)
    try:
        hlg_min_id /= total
    except ZeroDivisionError:
        hlg_min_id = 0
    return hlg_min_id, None


def hg_sim_calc_wpos(res, min_id, hlg2tot_con):
    """Calculate and return the average GCL, MMI, and MMP for all HGs shared
    within and HLG for each ome"""
    output_data = {}
    for hlg, gene_info in res.items():
        output_data[hlg] = {}
        tot_con = hlg2tot_con[hlg]
        for gene, data in gene_info.items():
            ome = gene[:gene.find('_')]
            ome_gcl = data[0]
            if not data[2]:
                try:
                    mi_p = [v[0] for v in data[1].values()]
                    ome_mi = min(mi_p)
                    mi_i = mi_p.index(ome_mi)
                    mp_p = [v[1] for v in data[1].values()]
                    ome_mp = mp_p[mi_i]
                except ValueError:
                    ome_mi, ome_mp = 0, 0
            else:
                ome_mi, ome_mp = 0, 0
    #            ome_mi, ome_mp = min_id, min_id
            if ome not in output_data[hlg]:
                output_data[hlg][ome] = (ome_gcl, ome_mi * tot_con, 
                                         ome_mp * tot_con)
            # if there's a paralog, take the best gcl
            else:
                if ome_gcl > output_data[hlg][ome][0]:
                    output_data[hlg][ome] = (ome_gcl, ome_mi * tot_con, 
                                             ome_mp * tot_con)
    return output_data


# deprecated
def hg_sim_calc_wopos(res, min_id, hlg2tot_con):
    """Calculate and return the average GCl and MMI for all HGs shared
    within an HLG for each ome, without considering MMP"""
    output_data = {}
    for hlg, gene_info in res.items():
        output_data[hlg] = {}
        tot_con = hlg2tot_con[hlg]
        for gene, data in gene_info.items():
            ome = gene[:gene.find('_')]
            ome_gcl = data[0]
            if not data[2]:
                try:
                    ome_mi = min(data[1].values())
                except ValueError:
                    ome_mi = 0
            else:
                ome_mi = 0
#                ome_mi = min_id
            if ome not in output_data[hlg]:
                output_data[hlg][ome] = (ome_gcl, ome_mi * tot_con)
            # if there's a paralog, take the best gcl
            else:
                if ome_gcl > output_data[hlg][ome][0]:
                    output_data[hlg][ome] = (ome_gcl, ome_mi * tot_con)
    return output_data


def hg_parse_and_calc(hg, hg_dict, hgx_dir, ome2i, min_id = 30, pos_data = True):
    """Manage the parsing and calculation of commitment to the locus and
    minimum similarity (identity and positives) for a given HG and the HLGs it
    is a part of"""

    # if there is percent positive data then use the functions that correspond
    # to that
    if pos_data:
       parse_func = parse_algn_wpos
       top_hits_func = get_top_hits_wpos
       hg_sim_func = hg_sim_calc_wpos
    # deprecated: lack of percent positives data [MMseqs]
    else:
       parse_func = parse_algn_wopos
       top_hits_func = get_top_hits_wopos
       hg_sim_func = hg_sim_calc_wopos

    gene2hlg = {}
    for hlg, genes in hg_dict.items():
        # ignore singletons
        if len(genes) > 1:
            for gene in genes:
                gene2hlg[gene] = hlg

    # grab the genes in an hg
    hg_genes = set(gene2hlg.keys())

    # parse the self-alignment results
    try:
        gene2algn = parse_func(hgx_dir, hg, hg_genes)
    except IndexError:
        if not pos_data:
            eprint(f'\tERROR: malformatted alignment: {hg}', flush = True)
            sys.exit(123)
        else:
            parse_func = parse_algn_wopos
            top_hits_func = get_top_hits_wopos
            hg_sim_func = hg_sim_calc_wopos
        try:
            gene2algn = parse_func(hgx_dir, hg, hg_genes)
        except IndexError:
            eprint(f'\tERROR: malformatted alignment {hg}', flush = True)
        return None, None, None, None
    except FileNotFoundError:
        return None, None, None, None

    # sort each query preferably by percent positives
    try:
        gene2algn = {
            k: sorted(v, key = lambda x: x[2], reverse = True) \
            for k,v in gene2algn.items()
            } 
    # otherwise run it by positives
    except IndexError:
        gene2algn = {k: sorted(v, key = lambda x: x[1], reverse = True) \
                     for k, v in gene2algn.items()}
    res = defaultdict(lambda: defaultdict(lambda: [0, {}, False]))
    todel_genes = defaultdict(lambda: defaultdict(list))

    # create a data structure that maintains all of the considered genes of an
    # HG in a particular HLG
    hlg2considered = {}
    for hlg, genes in hg_dict.items():
        failed_genes = True
        gene_set = set(genes)
        omes_set = set(x[:x.find('_')] for x in genes)

        # if there is a failed gene (too low identity to call a homolog, then
        # continually rerun until those have been exhaustively removed from
        # consideration
        while failed_genes:
            failed_genes = []
            # disregard the self from the length
            hgx_ome_len = len(omes_set) - 1
            if hgx_ome_len == 0:
                ome = genes[0][:genes[0].find('_')]
                todel_genes[hlg][ome].extend(genes)
                failed_genes.extend(genes)
                break
            # for the weighted average, we can start at this step
            # only consider unique omes because we are only taking the max
            considered_genes = len(set(x[:x.find('_')] for x in genes))
            hlg2considered[hlg] = considered_genes
            for gene in genes:
                ome = gene[:gene.find('_')] # identify the ome
#                if gene not in res[hlg]:
 #                   res[hlg][gene] = [0, {}, False]
        
                # to identify if the subject is in the family of omes
                hits = {gene}
                # while all cluster homologs aren't accounted for and there remain hits
                if gene not in gene2algn:
                    continue
                hits, res = top_hits_func(gene, hlg, gene_set, omes_set, hits, 
                                                     gene2algn, res, ome)
        
                # if there are missing hits (which means we are disregarding
                # missing alignments)
                if hgx_ome_len - len(res[hlg][gene][1]) > 0:
                    # then get the percent of observed hits that are valid
                    hit_len = res[hlg][gene][0] + len(res[hlg][gene][1])
                    try:
#* considered_genes \

                        res[hlg][gene][0] = len(res[hlg][gene][1]) / hit_len
                    except ZeroDivisionError:
#                        eprint(hlg, gene, gene_set, omes_set, res[hlg][gene], gene2algn, flush = True)
                        # no valid hits, the gene should not be called in the cluster
                        # typically happens from extension module
                        todel_genes[hlg][ome].append(gene)
                        failed_genes.append(gene)
                else:
                    # the GCL for this HG and ome is the total of the genes in this
                    # HG (disregarding self) divided by the total + the failed genes
#* considered_genes \

                    res[hlg][gene][0] = hgx_ome_len \
                                      / (hgx_ome_len + res[hlg][gene][0])
            # have to rerun the calculation if there are failed genes
            for failed_gene in failed_genes:
                gene_set.remove(failed_gene)
            genes = list(gene_set)
            omes_set = set(x[:x.find('_')] for x in genes)
   
    return hg, hg_sim_func(res, min_id, hlg2considered), \
           {hlg: dict(ome_dict) for hlg, ome_dict in todel_genes.items()}, \
           hlg2considered


def gcl_mngr(
    hgs, omes, hgx_dir, hgx2loc,
    db, gene2hg, clusplusminus, hg2gene, hlgs,
    hlg_omes, hlg_hgxs, ome2i, min_id, phylo,
    d2gcl, d2id_, d2pos, cpus = 1
    ):
    """Manage the calculation of GCL and similarity measurements (MMI and MMP)
    for each HLG"""
    i2ome = {v: k for k, v in ome2i.items()}

    # identify all the HGs that need to be parsed and organize them by their
    # HLG
    clus_hgs = {}
    for hlg, loci in hlgs.items():
        hlg_hgx = hlg_hgxs[hlg]
        omesc = hlg_omes[hlg]
        if hlg_hgx not in d2id_:
            clus_hgs[hlg] = [hlg_hgx, defaultdict(list)]
            hlg_hgx_set = set(hlg_hgx)
            for loc in loci:
                for gene in loc:
                    try:
                        hg = gene2hg[gene]
                    except KeyError:
                        continue
                    if hg in hlg_hgx_set:
                        clus_hgs[hlg][1][hg].append(gene)
        elif omesc not in d2id_[hlg_hgx]:
            clus_hgs[hlg] = [hlg_hgx, defaultdict(list)]
            hlg_hgx_set = set(hlg_hgx)
            for loc in loci:
                for gene in loc:
                    try:
                        hg = gene2hg[gene]
                    except KeyError:
                        continue
                    if hg in hlg_hgx_set:
                        clus_hgs[hlg][1][hg].append(gene)

            # {fam: [hgx, {hg: set(seqs)}}

    clus_hgs = {
        hlg: [d[0], {hg: set(v) for hg, v in d[1].items() if len(set(v)) > 1}] \
        for hlg, d in clus_hgs.items()
        } # make sets from it
    hg2genes = defaultdict(dict)
    for hlg, hgs in clus_hgs.items():
        for hg, gene_set in hgs[1].items():
            hg2genes[hg][hlg] = tuple(sorted(gene_set))

    # parse the alignments and acquire the preliminary quantitations of
    # similarity and GCL measurements on an HG-by-HG basis
    print('\tParsing alignments', flush = True)
    with mp.get_context('fork').Pool(processes = cpus) as pool:
        hg_results = pool.starmap(hg_parse_and_calc,
                              tqdm(((hg, hg_dict, hgx_dir, ome2i, min_id) \
                               for hg, hg_dict in hg2genes.items()),
                               total = len(hg2genes)))
        pool.close()
        pool.join()

    # prepare for quantifying the overall HLG GCL, MMI, MMP, and eventual CSB
    # by removing any loci and genomes that do not have sufficient
    # justification to warrant their retention following minimum similarity
    # thresholding
    print('\tQuantifying', flush = True)
    hlg_res = defaultdict(dict)
    deleted = 0
    hlg2tot_considered = defaultdict(int)
    hlgs = {hlg: list(locs) for hlg, locs in hlgs.items()}
    for hg, hg_res, todel_genes, hlg2considered in hg_results:
        todel = defaultdict(list)
        if hg_res:
            # acquire the total genes considered for each HLG
            for hlg, considered in hlg2considered.items():
                hlg2tot_considered[hlg] += considered
            for hlg, ome_dict in hg_res.items():
                hlg_res[hlg][hg] = []
                for ome, res in ome_dict.items():
                    hlg_res[hlg][hg].append(res)
            for hlg, ome_fails in todel_genes.items():
                for ome, failed_genes in ome_fails.items():
                    fails = set(failed_genes)
                    for i, loc in enumerate(hlgs[hlg]):
                        if loc[0][:loc[0].find('_')] == ome:
                            # if all/all but 1 of the genes failed, then delete it
                            if len([x for x in loc if x in fails]) >= len(loc) - 1:
                                todel[hlg].append(i)
            # remove any loci that now lack sufficient homology to justify them
            # being a part of an HLG, due to them have too low identity to be
            # considered
            for hlg, iz in todel.items():
                deleted += len(iz)
                for i in sorted(set(iz), reverse = True):
                    del hlgs[hlg][i]
    print(f'\t{deleted} loci removed due to lacking homology', flush = True)

    # delete any HLGs that have too few genomes to justify them
    todel = set()
    for hlg, locs in hlgs.items():
        omes = set(ome2i[loc[0][:loc[0].find('_')]] for loc in locs)
        if len(omes) > 2:        
            hlg_omes[hlg] = tuple(sorted(omes))
        else:
            todel.add(hlg)
    print(f'\t{len(todel)} HLGs removed for too few remaining omes', flush = True)
    for hlg in todel:
        del hlgs[hlg]
        del hlg_omes[hlg]
        del hlg_hgxs[hlg]
        
    # quantify the overall measurements for each HLG            
    hgx2omes2gcl = defaultdict(dict)
    hgx2omes2id = defaultdict(dict)
    hgx2omes2pos = defaultdict(dict)
    for hlg, hg_dict in tqdm(hlg_res.items(), total = len(hlg_res)):
        if hlg in todel:
            continue
        tot_considered = hlg2tot_considered[hlg]
        hgx, omes = hlg_hgxs[hlg], hlg_omes[hlg]
        gcls, ids, poss = [], [], []
        # calculate the average of each measurement for the overall HLG
        for hg, res in hg_dict.items():
            gcls.append(sum([x[0] for x in res]))
            ids.append(sum([x[1] for x in res])/len(res))
            try:
                poss.append(sum([x[2] for x in res])/len(res))
            except IndexError:
                poss.append(None)
            except TypeError:
                poss.append(None)
#        gcl = sum(gcls)/(len(gcls) * hlg2tot_considered[hlg])
        gcl = sum(gcls)/tot_considered
        id_ = sum(ids)/tot_considered
        try:
            pos = sum(poss)/tot_considered
        except ValueError:
            pos = None
        except TypeError:
            pos = None

        hgx2omes2gcl[hgx][omes] = gcl
        hgx2omes2id[hgx][omes] = id_
        if pos is not None:
            hgx2omes2pos[hgx][omes] = pos

    hgx2omes2pos = dict(hgx2omes2pos)
    return {**d2gcl, **hgx2omes2gcl}, \
        {**d2id_, **hgx2omes2id}, \
        {**d2pos, **hgx2omes2pos}, \
        {g: tuple(l) for g, l in hlgs.items()}, \
        {g: tuple(o) for g, o in hlg_omes.items()}


def find_missing_algns(hgs, hgx_dir):
    """Identify missing alignments that need to be run"""
    alnd_hgs = set([int(os.path.basename(x[:-4])) \
                    for x in collect_files(hgx_dir, 'out')])
    missing_alns = set(hgs).difference(alnd_hgs)
    return missing_alns


def parse_failures(fail_file):
    """Parse failed alignments file so that they are not continually rerun when
    it is unnecessary"""
    if os.path.isfile(fail_file):
        with open(fail_file, 'r') as raw:
            failed_hgs = [int(x.rstrip()) for x in raw]
    else:
        failed_hgs = []
    return failed_hgs


def prep_blast_cmds(db, hgs, hg_dir, hgx_dir, 
                    minid = 30, algorithm = 'diamond',
                    sensitivity = '', hg2gene = None,
                    cpus = 1, rerun = False):
    """Prepare commands for self-aligning HGs based on the inputted
    algorithm"""
    missing_alns = find_missing_algns(hgs, hgx_dir)

    # if failed runs are not to be rerun
    if not rerun:
        failed_hgs = parse_failures(hgx_dir + 'failed.txt')
        missing_alns = missing_alns.difference(set(failed_hgs))

    # prepare diamond commands
    if os.path.basename(algorithm) == 'diamond':
        dmnd_dbs = set([int(os.path.basename(x[:-5])) \
                    for x in collect_files(hgx_dir, 'dmnd')])
        missing_dmnds = set(missing_alns).difference(dmnd_dbs)
        cmds1 = [(algorithm, 'makedb', '--db', f'{hgx_dir}{hg}.dmnd',
                  '--in', f'{hg_dir}{hg}.faa', '--threads', '2') \
                  for hg in missing_dmnds]
        if sensitivity:
            cmds2 = [((algorithm, 'blastp', '--query', f'{hg_dir}{hg}.faa',
                      '--db', f'{hgx_dir}{hg}.dmnd', '-o', f'{hgx_dir}{hg}.out.tmp',
                      '--outfmt', '6', 'qseqid', 'sseqid', 'ppos', 'pident',
                      '--threads', '2', '--id', str(minid), '--ultra-sensitive',
                      '--no-self-hits', '--max-target-seqs', str(len(hg2gene[hg])), '&&'), 
                      ('mv', f'{hgx_dir}{hg}.out.tmp', f'{hgx_dir}{hg}.out')) \
                     for hg in missing_alns]
        else:
            cmds2 = [((algorithm, 'blastp', '--query', f'{hg_dir}{hg}.faa',
                      '--db', f'{hgx_dir}{hg}.dmnd', '-o', f'{hgx_dir}{hg}.out.tmp',
                      '--outfmt', '6', 'qseqid', 'sseqid', 'ppos', 'pident',
                      '--threads', '2', '--id', str(minid), '--max-target-seqs', 
                      str(len(hg2gene[hg])), '--no-self-hits', '&&'), 
                      ('mv', f'{hgx_dir}{hg}.out.tmp', f'{hgx_dir}{hg}.out')) \
                     for hg in missing_alns]
    # prepare BLASTp commands
    elif os.path.basename(algorithm) == 'blastp':
        cmds1 = []
        cmds2 = [((algorithm, '-query', f'{hg_dir}{hg}.faa', '-subject',
                   f'{hg_dir}{hg}.faa', '-out', f'{hgx_dir}{hg}.out.tmp',
                   '-outfmt', '6 "qseqid sseqid ppos pident"', '-num_threads',
                   str(cpus*2),
                   '-max_target_seqs', str(len(hg2gene[hg])), '&&'),
                   ('mv', f'{hgx_dir}{hg}.out.tmp', f'{hgx_dir}{hg}.out')) \
                  for hg in missing_alns]
    # deprecated - prepare MMseqs commands
    elif os.path.basename(algorithm) == 'mmseqs':
        mmseqs_dbs = set([int(os.path.basename(x[:-7])) \
                     for x in collect_files(hgx_dir, 'mmseqs')])
        missing_mmseqs = set(missing_alns).difference(mmseqs_dbs)
        cmds1 = [(algorithm , 'createdb', f'{hg_dir}{hg}.faa', f'{hgx_dir}{hg}.mmseqs',
                  '--createdb-mode', '1', '--shuffle', '0') for hg in missing_mmseqs]
        cmds2 = [((algorithm, 'search', f'{hgx_dir}{hg}.mmseqs', f'{hgx_dir}{hg}.mmseqs',
                  f'{hgx_dir}{hg}.raw', f'{hgx_dir}tmp{hg}', '--threads', str(cpus),
                  '--num-iterations', '3', '-s', '7.5', '-e', '0.001', 
                  '--max-seqs', str(len(hg2gene[hg])), '--max-rejected', '10', 
                  '--min-ungapped-score', '30', '&&'),
                  (algorithm, 'filterdb', f'{hgx_dir}{hg}.raw', f'{hgx_dir}{hg}.raw.filter', 
                   '--comparison-operator', 'ge', '--comparison-value', str(minid/100), 
                   '--filter-column', '2', '--threads', str(cpus), '&&'),
                  (algorithm, 'convertalis', f'{hgx_dir}{hg}.mmseqs', f'{hgx_dir}{hg}.mmseqs',
                   f'{hgx_dir}{hg}.raw.filter', f'{hgx_dir}{hg}.out.tmp', '--format-output', 
                   'query,target,pident', '--threads', str(cpus), '&&'),
                  ('mv', f'{hgx_dir}{hg}.out.tmp', f'{hgx_dir}{hg}.out', '&&'),
                  ('rm', '-rf', f'{hgx_dir}tmp{hg}', f'{hgx_dir}{hg}.raw', 
                   f'{hgx_dir}{hg}.raw.filter')) for hg in missing_alns]

    return cmds1, cmds2


def run_blast(hgs, db, hg_dir, hgx_dir, algorithm = 'diamond', 
              printexit = False, sensitivity = '', hg2gene = None,
              skipalgn = False, fallback = False, minid = 30, cpus = 1):
    """Run HG self-alignment commands"""
#    hgs = sorted(set(chain(*list(hgx2loc.keys()))))
    db_cmds, algn_cmds = prep_blast_cmds(db, hgs, hg_dir, hgx_dir, 
                                         minid = minid, rerun = not skipalgn,
                                         algorithm = algorithm, 
                                         sensitivity = sensitivity,
                                         hg2gene = hg2gene, cpus = cpus)
    if algn_cmds:
        # make the databases if necessary (Diamond)
        if db_cmds:
            with open(hgx_dir + '../makedb.sh', 'w') as out:
                out.write('\n'.join([' '.join(x) for x in db_cmds]))
        # write and run the search commands
        with open(hgx_dir + '../srch.sh', 'w') as out:
            if len(algn_cmds[0]) == 2:
                for algn, tmp_mv in algn_cmds:
                    out.write(' '.join(algn) + '\n' + ' '.join(tmp_mv) + '\n')
            else:
                for algn, fltr, convert, tmp_mv, rm in algn_cmds:
                    out.write(' '.join(algn[:-1]) + ' && \n ' \
                            + ' '.join(fltr[:-1]) + ' && \n ' \
                            + ' '.join(convert[:-1]) + ' && \n ' \
                            + ' '.join(tmp_mv[:-1]) + ' && \n ' \
                            + ' '.join(rm) + ' \n\n')
        # exit once the commands are outputted to allow for user-controlled
        # parallelization
        if printexit:
            print(f'\n{algorithm} db commands outputted to ' \
                 + f'{hgx_dir}../makedb.sh; run this first', flush = True)
            print(f'\n{algorithm} commands outputted to {hgx_dir}../srch.sh',
                  flush = True)
            sys.exit(0)

        # NEED to rewrite out what skipalgn does
        elif not skipalgn:
            if db_cmds:
                print(f'\tBuilding {len(db_cmds)} aligner DBs', flush = True)
                multisub(db_cmds, verbose = 2, processes = cpus)
            print(f'\tAligning {len(algn_cmds)} HGs', flush = True)
            if algorithm == 'diamond':
                multisub(algn_cmds, verbose = 2, processes = cpus,
                         injectable = True)
            else: # leave the multithreading optimization to the program
    #            multisub(algn_cmds, verbose = 2, processes = round((cpus - 1)/2),
     #                    injectable = True)
                # launch 1 subprocess to minimize python overhead
                subprocess.run([os.environ['SHELL'], f'{hgx_dir}../srch.sh'],
                                stdout = subprocess.DEVNULL)

    # identify missing alignments
    missing_alns = find_missing_algns(hgs, hgx_dir)
    # call these alignments failed
    if not fallback:
        print(f'\t\t{len(missing_alns)} failed alignments will be skipped', 
              flush = True)
    # if there are missing alignments and a fallback method, then run that
    elif missing_alns:
        print(f'\t\t{len(missing_alns)} failed alignments will fallback to diamond', 
              flush = True)
        db_cmds, algn_cmds = prep_blast_cmds(db, hgs, hg_dir, hgx_dir, 
                                         minid = minid, rerun = True,
                                         algorithm = 'diamond', 
                                         sensitivity = None,
                                         hg2gene = hg2gene, cpus = cpus)
        if db_cmds:
            print(f'\tBuilding {len(db_cmds)} diamond DBs', flush = True)
            multisub(db_cmds, verbose = 2, processes = cpus)
        print(f'\tAligning {len(algn_cmds)} HGs', flush = True)
        multisub(algn_cmds, verbose = 2, processes = cpus,
                 injectable = True)
        missing_alns = find_missing_algns(hgs, hgx_dir)

    # finally, output the failed alignments for future reruns to reference 
    with open(hgx_dir + 'failed.txt', 'w') as out:
        out.write('\n'.join([str(x) for x in missing_alns]))


def gcl_main(
    hgx2loc, wrk_dir, ome2i, hg_dir, hgx_dir,
    algorithm, db, gene2hg, plusminus, hg2gene, phylo,
    old_path = 'gcl.pickle',
    hlgs = None, hlg_hgxs = None,
    hlg_omes = None, hlg2clan = {}, 
    cpus = 1, printexit = False,
    algn_sens = '', skipalgn = False, minid = 30,
    fallback = False
    ):
    """The main function for controlling and executing the management of
    running GCL, MMI, MMP, and CSB quantitation"""

    # determine if the hgx directory exists, and unzip if necessary
    if not checkdir(hgx_dir, unzip = True, rm = True):
        os.mkdir(hgx_dir)

    # open existing data structures to rerun from what is already complete
    if os.path.isfile(wrk_dir + old_path):
        print('\tLoading previous GCL, MMI, and MMP results', flush = True)
        with open(wrk_dir + 'gcl.pickle', 'rb') as pickin:
            d2gcl = pickle.load(pickin)
        with open(wrk_dir + 'mmi.pickle', 'rb') as pickin:
            d2id_ = pickle.load(pickin)
        if os.path.isfile(wrk_dir + 'mmp.pickle'):
            with open(wrk_dir + 'mmp.pickle', 'rb') as pickin:
                d2pos = pickle.load(pickin)
        else:
            d2pos = {}
    else:
        d2gcl, d2id_, d2pos = {}, {}, {}

    # output fastas for any HGs that do not have them
    # NEED to figure out why this is necessary now when earlier steps should
    # have already completed this
    hgs = list(chain(*list(hlg_hgxs.values())))
    hg_dir = hg_fa_mngr(wrk_dir, None, hgs, 
                        db, hg2gene, cpus = cpus,
                        low_mem = True)

    # run HG self-alignment to set the stage for similarity calculations
    run_blast(hgs, db, hg_dir, hgx_dir, #minid = minid, let minid be at locus sim step
              algorithm = algorithm, printexit = printexit,
              sensitivity = algn_sens, hg2gene = hg2gene,
              skipalgn = skipalgn, fallback = fallback,
              cpus = cpus)

    # calculate the proxies of coordinated gene evolution
    d2gcl, d2id_, d2pos, hlgs, hlg_omes = gcl_mngr(
        list(hgs), list(ome2i.keys()), hgx_dir, hgx2loc,
        db, gene2hg, plusminus, hg2gene, hlgs,
        hlg_omes, hlg_hgxs, ome2i, minid, phylo,
        d2gcl, d2id_, d2pos, cpus = cpus #d2pos, cpus = cpus
        )
#    hgx_dirTar = mp.Process(target=tardir, args=(hgx_dir, True))
 #   hgx_dirTar.start() # when to join ...

    # output the data structures for future swift parsing and update the HLG
    # data based on any HLGs that were removed following determination of being
    # below minimum gene similarity thresholds
    with open(f'{wrk_dir}hlgs.pickle', 'wb') as out:
        pickle.dump([hlgs, hlg_omes, hlg_hgxs, hlg2clan], out)
    with open(wrk_dir + 'gcl.pickle', 'wb') as pickout:
        pickle.dump(d2gcl, pickout)
    with open(wrk_dir + 'mmi.pickle', 'wb') as pickout:
        pickle.dump(d2id_, pickout)
    if d2pos:
        with open(wrk_dir + 'mmp.pickle', 'wb') as pickout:
            pickle.dump(d2pos, pickout)

    return d2gcl, d2id_, d2pos, hlgs, hlg_omes



import os
import re
import gzip
import shutil
import pickle
import subprocess
import multiprocessing as mp
import plotly.graph_objects as go
from math import log
from tqdm import tqdm
from itertools import chain
from collections import defaultdict
from plotly.subplots import make_subplots
from mycotools.acc2gff import grab_gff_acc
from mycotools.lib.kontools import eprint
from mycotools.lib.biotools import list2gff, gff2list, fa2dict, dict2fa
from cloci.lib import input_parsing

# NEED to add PCA output
# NEED locID output and clans
# NEED gbk output option
# NEED to add merge abutting clusters feature
# NEED annotation for gcfs resume and hlgs independent

def mk_subplots(labels, axes, axes_labels, alpha = 0.6):
    
    colors = [
        [240,163,255], [0,117,220], [153,63,0], [76,0,92], [25,25,25],
        [0,92,49], [43,206,72], [255,204,153], [128,128,128], [148,255,181],
        [143,124,0], [157,204,0], [194,0,136], [0,51,128], [255,164,5],
        [255,168,187], [66,102,0], [255,0,16], [94,241,242], [0,153,143],
        [224,255,102], [116,10,255], [153,0,0], [255,255,128], [255,255,0], [255,80,5]
        ]
#    random.shuffle(colors)
    pl_clr = [
        'rgba(' + ', '.join([str(y) for y in x]) + ', ' + str(alpha) + ')' \
        for x in colors
        ]
    fig = make_subplots(
        rows=len(axes) - 1, cols=len(axes) - 1,
        start_cell = "bottom-left"
        )

    clr_count = 0
    for i0, axis0 in enumerate(axes):
        for i1, axis1 in enumerate(axes[i0+1:]):
            fig.add_trace(
                go.Scatter(
                    mode = 'markers', x=axis0, y=axis1,
                    marker = dict(
                        color = pl_clr[clr_count],
                        size = 5,
                        opacity = alpha
                        ),
                    text = labels
                    ), row = i0+1, col = i0+i1+1,
                )
            fig.update_xaxes(
                title_text=axes_labels[i0], row = i0+1, col =
                i0+i1+1
                )
            fig.update_yaxes( 
                title_text=axes_labels[i0+i1+1], row = i0+1, col =
                i0+i1+1
                )
            clr_count += 1
    
    return fig 


def mk_3d(labels, axes, axes_labels, alpha = 0.6):

    fig = go.Figure()
    fig.add_trace(go.Scatter3d(
        mode = 'markers',
        x = axes[0], y = axes[1], z = axes[2],
        marker = dict(
            color = 'rgba(163, 25, 169,' + str(alpha) + ')',
            size = 5, opacity = alpha
            ),
        text = labels, showlegend = False
        )
    )
    fig.update_layout(scene = dict(
                    xaxis_title=axes_labels[0],
                    yaxis_title=axes_labels[1],
                    zaxis_title=axes_labels[2]))#,
#                    width=700,
 #                   margin=dict(r=20, b=10, l=10, t=10))

    return fig


def extract_sig_clus(
    clus_scores_dict, hgx2loc, top_hgxs, ome2i
    ):
    """Create a hash to seed retrieving clusters based on
    their entry in hgx2loc."""
        
    sig_clus = defaultdict(dict)
    for top in top_hgxs:
        hgx, gcf, omeIs = top[0], top[1], top[2]
        for gene in hgx2loc[hgx]:
            ome = gene[:gene.find('_')]
            if ome2i[ome] in omeIs:
                if gene not in sig_clus[ome]:            
                    sig_clus[ome][gene] = [[set(hgx), gcf]]
                else:
                    sig_clus[ome][gene].append([set(hgx), gcf])
    
    return dict(sig_clus)


def hash_clusters(gff_path):
    """sliding window of size clusplusminus, identify sequences that may have
    hgs that may be significant, grab their window, and check for a significant
    higher order og combo in the locus"""
    gff_list = gff2list(gff_path)
    cds_dict = input_parsing.compileCDS(gff_list, os.path.basename(gff_path).replace('.gff3',''))
    return cds_dict


def merge_clusters(clus_out, cds_dict):
    """because we have significant clusters with borders, we can merge each
    cluster and assume that at a certain number of hgs in a combo will never
    happen via random sampling; therefore, their significance in the
    microsynteny adjustment is sufficient for justification for merging"""
    
    comp_clus, change = defaultdict(list), False
    for scaf, clusters in clus_out.items():
        gene_list = cds_dict[scaf]
        while clusters: # while there are clusters
            any_intersect, toDel = False, [0] # we haven't found intersection
            clus0 = clusters[0] # grab the first tuple
            loc0 = clus0[0] # grab the genes in the cluster
            low_i0 = gene_list.index(loc0[0])
            hi_i0 = gene_list.index(loc0[-1])
            for i, clus1 in enumerate(clusters[1:]): # look at all other clusters
                loc1 = clus1[0] # grab their loci
                low_i1 = gene_list.index(loc1[0])
                hi_i1 = gene_list.index(loc1[-1])
                if low_i1 - hi_i0 == 1 and False:
                    any_intersect, change = True, True # we found overlap in
                    # clus_out and in this specific cluster comparison
 #                   hgx = tuple(sorted(list(set(clus0[0]).union(set(clus1[0])))))
                    # obtain the higher order og combination as the union of the
                    # two cluster sets, then sort it and store as a tuple
                    comp_clus[scaf].append([
#                        hgx, 
                        loc0 + loc1, 
                        clus0[1].union(clus1[1]) #, merge_x
                        ])
                    # append the og combo and the union of the two loci's genes
                    # protein order is random
                    toDel.append(i) # we can delete the intersected locus as well
                    # as the first
                elif low_i0 - hi_i1 == 1 and False:
                    any_intersect, change = True, True
                    comp_clus[scaf].append([loc1 + loc0, clus0[1].union(clus1[1])])
                    toDel.append(i)
                elif set(loc0).intersection(set(loc1)) and False:
                    any_intersect, change = True, True
                    comp_clus[scaf].append([set(loc0).union(set(loc1)), clus0[1].union(clus1[1])])
                    comp_clus[scaf][-1][0] = sorted(comp_clus[scaf][-1][0], 
                                                    key = cds_dict[scaf].index)   
                    toDel.append(i)
            if not any_intersect: # if there is not any intersection in the locus
                comp_clus[scaf].append(clus0) #, clus0[2]]) # simply append the original
                # result
            toDel.sort(reverse = True) # sort the todel from highest to lowest to
            # not mess up order when indexing
            for i in toDel:
                clusters.pop(i)
            
    return comp_clus, change


def clus2scaf(cds_dict, clusters):

    gene2scaf, index_map = {}, {}
    for scaf, genes in cds_dict.items():
        index_map[scaf] = {}
        for i, gene in enumerate(genes):
            index_map[scaf][gene] = i
            gene2scaf[gene] = scaf

    clus_out = defaultdict(list)
    for loc, gcf in clusters:
        gene0 = loc[0]
        scaf = gene2scaf[gene0]
        ord_loc = sorted(loc, key = lambda pair: index_map[scaf][pair])
#        hgs = []
 #       for gene in ord_loc:
  #          try:
   #             hgs.append(gene2hg[gene])
    #        except KeyError:
     #           hgs.append('')
      #  clus_info = [hgs, ord_loc, {gcf}]
        
        clus_out[scaf].append([ord_loc, gcf])

    sorted_clus_out = {k: sorted(v, key = lambda x: index_map[k][x[0][0]]) \
                       for k, v in clus_out.items()}

    return sorted_clus_out, index_map


def write_clusters(clusters, ome, out_file, gff_path, gene2hg, clusplusminus = 10):
    cds_dict = input_parsing.compileCDS(gff2list(gff_path), 
                                        os.path.basename(gff_path).replace('.gff3',''))

    clus_out, scaf2gene2i = clus2scaf(cds_dict, clusters)
    merge_out = defaultdict(list)
    for scaf, locs in clus_out.items():
        for loc, gcf in locs:
            merge_out[scaf].append([loc, {gcf}])
  #  change = True
#    while change: # so long as there is a single merge
 #       clus_out, change = merge_clusters(clus_out, cds_dict) # attempt to merge clusters
#    clus_out = sorted(clus_out, key = lambda x: len(x[0]), reverse = True)
    # sort the clusters by the size of the OG combination tuple highest to
    # lowest

    for scaf, clusters in merge_out.items():
        mapping = {v: i for i, v in enumerate(cds_dict[scaf])}
        clusters = sorted(clusters, key = lambda pair: mapping[pair[0][0]])
        for i, clus in enumerate(clusters):
            clus[0] = sorted(clus[0], key = cds_dict[scaf].index)
            clusHGs = []
            for gene in clus[0]:
                try:
                    clusHGs.append(gene2hg[gene])
                except KeyError: # gene not in an OG/is a singleton
                    clusHGs.append('')
            clus.append(clusHGs)
   #         scaf = clus[2]
    #        scaf = clus[3][:20]
     #       if clus[3] != scaf:
    #            scaf = clus[3][:20] + '..'
#            count = 0
 #           while scaf + '_' + str(count) in scafs:
  #              count += 1
            name = scaf + '_' + str(i)
#            scafs.add(name)
            clus.append(name)
        clus_out[scaf] = clusters
    

    out_genes = []
    with open(out_file, 'w') as out:
        out.write('#name\thgs\tgenes\tgcfs\n')
        for scaf, clusters in clus_out.items():
            for clus in clusters:
 #           clus[2] = sorted(clus[2])
                hgs = ','.join([str(x) for x in clus[2]]) # og1,og2,...ogn
                genes = ','.join(clus[0]) # prot1,prot2,prot3,...protn
                gcfs = ';'.join([str(x) for x in sorted(clus[1])])
                out.write(clus[3] + '\t' + hgs + '\t' + genes + '\t' + gcfs + '\n')
                out_genes.append(clus[0])

    return ome, out_genes


def read_ome_hlgs(ome, ome_hlg_file):
    out_genes = []
    try:
        with open(ome_hlg_file, 'r') as raw:
            for line in raw:
                if not line.startswith('#'):
#        out.write('#name\thgs\tgenes\tgcfs\n')
                    data = line.rstrip().split()
                    genes = data[2].split(',')
                    out_genes.append(genes)
        return ome, out_genes
    except FileNotFoundError:
        return ome, []




def output_ann_fas(ann_dir, ome, ome_genes, prot_path):
    faa = fa2dict(prot_path)
    genes = list(chain(*ome_genes))
    try:
        out_faa = {k: faa[k] for k in genes}
        with open(ann_dir + ome + '.faa', 'w') as out:
            out.write(dict2fa(out_faa))
        return ome, True
    except KeyError:
        eprint('\t\tERROR: ' + ome + ' accessions modified from input',
               flush = True)
        return ome, False


def output_res_faas(genes_in, prot_paths, ann_dir, cpus = 1):
    with mp.Pool(processes = cpus) as pool:
        res = pool.starmap(output_ann_fas, 
                           tqdm(((ann_dir, ome, ome_genes, prot_paths[ome]) \
                                 for ome, ome_genes in genes_in.items()),
                                total = len(genes_in)))

    return set(ome for ome, passing in res if not passing)

def call_hmmsearch(ome, pfam, ann_dir, faa_file, threads = 4):
    out_file = ann_dir + ome + '_pfam.out'
    if not os.path.isfile(out_file):
        hmmsearch = subprocess.call([
            'hmmsearch', '--cpu', str(threads), 
            '--domtblout', out_file + '.tmp', pfam, faa_file
            ], stdout = subprocess.DEVNULL, stderr = subprocess.DEVNULL
            )
        os.rename(out_file + '.tmp', out_file)
    else:
        hmmsearch = 0
    return ome, hmmsearch


def run_hmmsearch(pfam, passing_omes, ann_dir, cpus, 
                  db = None, threads = 4):
    """Manage running hmmsearch for Pfam annotations. Default runs just the
    Pfam hits, inputting a db runs for the full genome"""

    with mp.Pool(processes = round(cpus/2)) as pool:
        if db:
            hmm_res = pool.starmap(call_hmmsearch, 
                                   tqdm(((ome, pfam, ann_dir,
                                     db[ome]['faa'], threads) \
                                     for ome in list(passing_omes)),
                                    total = len(passing_omes)))
        else:
            hmm_res = pool.starmap(call_hmmsearch, 
                               tqdm(((ome, pfam, ann_dir,
                                     f'{ann_dir}{ome}.faa', threads) \
                                     for ome in list(passing_omes)),
                                    total = len(passing_omes)))

    failed_omes = []
    for ome, exit_code in hmm_res:
        if exit_code:
            eprint('\t\tERROR: ' + ome + ' failed hmmsearch', flush = True)
            failed_omes.append(ome)

    return failed_omes

def call_iprscan(ome, iprscan, ann_dir, threads = 4):
    out_file = ann_dir + ome + '_ipr.out'
    faa_file = ann_dir + ome + '.faa'
    if not os.path.isfile(out_file):
        ipr_code = subprocess.call([
            iprscan, '-i', faa_file, '-cpu', str(threads), 
            '-o', out_file + '.tmp', '-f', 'TSV'
            ], stdout = subprocess.DEVNULL, stderr = subprocess.DEVNULL
            )
    else:
        ipr_code = 0
    return ome, ipr_code 


def run_ipr_scan(iprscan, passing_omes, ann_dir, cpus):
    """Initialize IPR scan for all genomes"""
    threads = 4
    with mp.Pool(processes = round(cpus/2)) as pool:
        ipr_res = pool.starmap(call_iprscan, 
                               tqdm(((ome, iprscan, ann_dir, threads) \
                                     for ome in list(passing_omes)),
                                    total = len(passing_omes)))

    failed_omes = []
    for ome, exit_code in ipr_res:
        if exit_code:
            eprint('\t\tERROR: ' + ome + ' failed InterProScan', flush = True)
            failed_omes.append(ome)

    return failed_omes




def parse_hmm_res(ome, ann_dir, evalue, threshold, max_hits = None):
    ome_out = ann_dir + ome + '_pfam.out'
#    lineComp = re.compile(r'([^ ]+)')
    res = defaultdict(list)
    with open(ome_out, 'r') as raw:
        for line in raw:
            if not line.startswith('#'):
#                data = lineComp.findall(line.rstrip())
                data = line.rstrip().split()
                hit_perc = (int(data[18]) - int(float(data[17])))/int(data[5])
                if hit_perc > threshold:
                    if float(data[6]) < evalue: # and float(data[7]) > score:
                        # pfam, name, cov_perc, evalue, bit, algn_start, algn_end
                        res[data[0]].append(( 
                            data[4], data[3], hit_perc,
                            float(data[21]), int(data[17]), int(data[18])
                            ))
    ome_res = {}
    for gene, gene_hits in res.items():
        gene_hits = sorted(gene_hits, key = lambda x: x[4], reverse = True)
        if max_hits:
            gene_hits = gene_hits[:max_hits]
        todel, hits = [], set()
        for i, hit in enumerate(gene_hits):
            if hit[0] in hits:
                todel.append(i)
            hits.add(hit)
        for i in reversed(todel):
            del gene_hits[i]
        ome_res[gene] = gene_hits

    return ome, ome_res


def compile_hmm_res(ann_dir, passing_omes, evalue = 0.001, 
                    threshold = 0.5, cpus = 1, max_hits = None):
        
    with mp.Pool(processes = cpus) as pool:
        omebyome_res = pool.starmap(parse_hmm_res, 
                               tqdm(((ome, ann_dir, evalue, threshold) \
                                     for ome in list(passing_omes)), 
                                    total = len(passing_omes)))
    ome_res = {}
    for ome, res in omebyome_res:
        ome_res[ome] = res

    return ome_res


def parse_ipr_res(ome, ann_dir, evalue, threshold):
    # NOT BUILT YET
    ome_out = ann_dir + ome + '_ipr.out'
    lineComp, res = re.compile(r'([^ ]+)'), defaultdict(list)
    with open(ome_out, 'r') as raw:
        for line in raw:
            if not line.startswith('#'):
                data = lineComp.findall(line.rstrip())
                hit_perc = (int(data[15]) - int(float(data[14])))/int(data[5])
                if hit_perc > threshold:
                    if float(data[6]) < evalue: # and float(data[7]) > score:
                        # pfam, name, cov_perc, evalue, bit, algn_start, algn_end
                        res[data[0]].append(( 
                            data[4], data[3], hit_perc, float(data[6]), 
                            float(data[7]), int(data[16]), int(data[17])
                            ))
    ome_res = {}
    for gene, gene_hits in res.items():
        gene_hits = sorted(gene_hits, key = lambda x: x[3], reverse = True)
        todel, hits = [], set()
        for i, hit in enumerate(gene_hits):
            if hit[0] in hits:
                todel.append(i)
            hits.add(hit)
        for i in reversed(todel):
            del gene_hits[i]
        ome_res[gene] = gene_hits

    return ome, ome_res




def compile_ipr_res(ann_dir, passing_omes, evalue = 0.01, threshold = 0.5, cpus = 1):
        
    with mp.Pool(processes = cpus) as pool:
        omebyome_res = pool.starmap(parse_ipr_res, 
                               tqdm(((ome, ann_dir, evalue, threshold) \
                                     for ome in list(passing_omes)), 
                                    total = len(passing_omes)))
    ome_res = {}
    for ome, res in omebyome_res:
        ome_res[ome] = res

    return ome_res




def ann_mngr(genes_list, prot_paths, wrk_dir, pfam, ipr_scan,
             evalue = 0.0001, threshold = 0.5, cpus = 1):

    ann_res = None
    ann_dir = wrk_dir + 'ann/'
    if not os.path.isdir(ann_dir):
        os.mkdir(ann_dir)

    failed_omes = output_res_faas(genes_list, prot_paths, ann_dir, cpus = cpus)
    passing_omes = set(genes_list.keys()).difference(failed_omes)

    if pfam:
        print("\tHmmsearch'ing Pfam database", flush = True)
        hmm_failed_omes = run_hmmsearch(pfam, passing_omes, ann_dir, cpus)
        passing_omes = passing_omes.difference(set(hmm_failed_omes))
        failed_omes = failed_omes.union(set(hmm_failed_omes))
        print('\tParsing hmmsearch output', flush = True)
        ann_res = compile_hmm_res(ann_dir, passing_omes, evalue, threshold, cpus)
    elif ipr_scan:
        print('\tRunning InterProScan', flush = True)
        ipr_failed_omes = run_ipr_scan(ipr_scan, passing_omes, ann_dir, cpus)
        passing_omes = passing_omes.difference(set(ipr_failed_omes))
        failed_omes = failed_omes.union(set(ipr_failed_omes))
        print('\tParsing InterProScan output', flush = True)
        pass

    return ann_res, failed_omes


def annotate_clusters(genes_list, #gff_path, prot_path, 
                      ome, ome_dir, gene2hg, prefix = 'hlg',
                      ann_res = {}):

    if not os.path.isfile(f'{ome_dir}{ome}/{prefix}.tsv'):
        return

    with open(ome_dir + ome + f'/{prefix}.tsv.tmp', 'w') as out:
        out.write('#name\thgs\tgenes\tgcfs\tpfams\n')
        with open(ome_dir + ome + f'/{prefix}.tsv', 'r') as raw:
            for line in raw:
                if not line.startswith('#'):
                    clus_d = line.split()
                    genes = clus_d[2].split(',')
                    anns_p = [ann_res[x] if x in ann_res else None for x in genes]
                    anns = ['|'.join([y[1] if y else '' for y in x]) \
                            if x else '' for x in anns_p]
                    ann_str = ';'.join(anns)
                    out.write('\t'.join(clus_d + [ann_str]) + '\n')
    shutil.move(f'{ome_dir}{ome}/{prefix}.tsv.tmp', 
                f'{ome_dir}{ome}/{prefix}.tsv')

#    return ome#, svg_dict
#    return ome, clus_gffs, clus_fas, svg_dict


def output_hgxs(hgx2dist, hgx2omes, hgx2i, i2ome, out_dir):
    maxhgxd = max(hgx2dist.values())
    minhgxd = min(hgx2dist.values())

    denom = maxhgxd - minhgxd
    hgx_output = []
    for hgx, omes in hgx2omes.items():
        hgx_id = hgx2i[hgx]
        hgx_output.append([
            ','.join([str(x) for x in hgx]), hgx_id,
            (log(hgx2dist[hgx]) - minhgxd)/denom,
            ','.join([i2ome[x] for x in omes])
            ]) # HGxs at this stage are not segregated into groups
    hgx_output = sorted(hgx_output, key = lambda x: x[3], reverse = True)
    with gzip.open(out_dir + 'hgxs.tsv.gz', 'wt') as out:
        out.write('#hgs\thgx_id\tnrm_log_tmd\tomes')#\tpdd\tomes')
        for entry in hgx_output:
            out.write('\n' + '\t'.join([str(x) for x in entry]))

def write_hlgs_txt_wpos(hlg_hgxs, hlg_omes, logg2d, 
                        hlg2clan, omes2patch, 
                        hgx2omes2gcl, hgx2omes2id, hgx2omes2pos, i2ome,
                        out_dir, group_name):
    hlg_output = []
    for i, omesc in hlg_omes.items():
        hlg_hgx = hlg_hgxs[i]
        try:
            hlg_output.append([
                ','.join([str(x) for x in hlg_hgx]), i, hlg2clan[i],
                logg2d[omesc], omes2patch[omesc], 
                hgx2omes2gcl[hlg_hgx][omesc],
                hgx2omes2id[hlg_hgx][omesc], hgx2omes2pos[hlg_hgx][omesc],
                100 * (1 - (hgx2omes2id[hlg_hgx][omesc]/hgx2omes2pos[hlg_hgx][omesc])), 
                ','.join([str(i2ome[x]) for x in omesc])#,
         #        dnds_dict[hgx][0], dnds_dict[hgx][1], str(dnds_dict[hgx][2]),
                ])
        except ZeroDivisionError:
            hlg_output.append([
                ','.join([str(x) for x in hlg_hgx]), i, hlg2clan[i],
                logg2d[omesc], omes2patch[omesc], 
                hgx2omes2gcl[hlg_hgx][omesc],
                hgx2omes2id[hlg_hgx][omesc], hgx2omes2pos[hlg_hgx][omesc],
                0, ','.join([str(i2ome[x]) for x in omesc])#,
         #        dnds_dict[hgx][0], dnds_dict[hgx][1], str(dnds_dict[hgx][2]),
                ])
        except KeyError:
            try:
                hlg_output.append([
                   ','.join([str(x) for x in hlg_hgx]), i, hlg2clan[i],
                   logg2d[omesc], omes2patch[omesc], 
                   hgx2omes2gcl[hlg_hgx][omesc],
                   hgx2omes2id[hlg_hgx][omesc], 'na', 'na',
                   ','.join([str(i2ome[x]) for x in omesc])#,
            #        dnds_dict[hgx][0], dnds_dict[hgx][1], str(dnds_dict[hgx][2]),
                   ])
            except KeyError:
                eprint(f'ERROR: missing HGx/omes from results: {hlg_hgx}, {omesc}', flush =  True)

    hlg_output = sorted(hlg_output, key = lambda x: x[3], reverse = True)
    with gzip.open(out_dir + f'{group_name}s.tsv.gz', 'wt') as out:
        out.write(f'#hgs\t{group_name}\thlc\tnrm_log_tmd\tpds' \
                + '\tgcl\tmmi\tmmp\tcsb\tomes') #\tmmp\tomes') #+ \
            #'selection_coef\tmean_dnds\tog_dnds\t' + \
         #   'total_dist'
#            )
        for entry in hlg_output:
            out.write('\n' + '\t'.join([str(x) for x in entry]))
    return hlg_output

def write_hlgs_txt_wopos(hlg_hgxs, hlg_omes, logg2d, 
                        hlg2clan, omes2patch, 
                        hgx2omes2gcl, hgx2omes2id, i2ome,
                        out_dir, group_name):
    hlg_output = []
    for i, hlg_hgx in hlg_hgxs.items():
        omesc = hlg_omes[i]
        hlg_output.append([
            ','.join([str(x) for x in hlg_hgx]), i, hlg2clan[i],
            logg2d[omesc], omes2patch[omesc], 
            hgx2omes2gcl[hlg_hgx][omesc],
            hgx2omes2id[hlg_hgx][omesc],
            ','.join([str(i2ome[x]) for x in omesc])#,
     #        dnds_dict[hgx][0], dnds_dict[hgx][1], str(dnds_dict[hgx][2]),
            ])

    hlg_output = sorted(hlg_output, key = lambda x: x[3], reverse = True)
    with gzip.open(out_dir + f'{group_name}s.tsv.gz', 'wt') as out:
        out.write(f'#hgs\t{group_name}\thlc\tnrm_log_tmd\tpds' \
                + '\tgcl\tmmi\tomes') #\tmmp\tomes') #+ \
            #'selection_coef\tmean_dnds\tog_dnds\t' + \
         #   'total_dist'
#            )
        for entry in hlg_output:
            out.write('\n' + '\t'.join([str(x) for x in entry]))
    return hlg_output


def calc_logg2d(hlg_omes, omes2dist):
    logg2d_prep = {}
    for omesc in hlg_omes.values():
        logg2d_prep[omesc] = log(omes2dist[omesc])
    maxhlgd = max(logg2d_prep.values()) # max observed, even of those truncated/removed
    minhlgd = min(logg2d_prep.values())
    denom = maxhlgd - minhlgd
    logg2d = {k: (v - minhlgd)/denom for k, v in logg2d_prep.items()}
    return logg2d


def threshold_hlg_by_dist(logg2d, dist_thresh, hlgs, hlg_hgxs, hlg_omes, omes2dist):
    if dist_thresh:
        gcfs, gcf_omes, gcf_hgxs = {}, {}, {}
        for hlg, locs in hlgs.items():
            hgxc = hlg_hgxs[hlg]
            omesc = hlg_omes[hlg]
            if logg2d[omesc] >= dist_thresh:
                gcfs[hlg] = locs
                gcf_omes[hlg] = omesc
                gcf_hgxs[hlg] = hgxc
        return gcfs, gcf_omes, gcf_hgxs
    else:
        return hlgs, hlg_omes, hlg_hgxs


def threshold_hlg_wpos(gcl_thresh, patch_thresh, id_perc, pos_perc, csb_thresh, hlgs,
                       hlg_hgxs, hlg_omes, omes2patch, hgx2omes2gcl,
                       hgx2omes2id, hgx2omes2pos):

    if any(x > 0 for x in [gcl_thresh, patch_thresh, id_perc, pos_perc, csb_thresh]):
        gcfs, gcf_omes, gcf_hgxs = {}, {}, {}
        for hlg, locs in hlgs.items():
            check = False # have we added a new list
            hgxc = hlg_hgxs[hlg]
            omesc = hlg_omes[hlg]
            try:
                patch = omes2patch[omesc]
                gcl = hgx2omes2gcl[hgxc][omesc]
                id_, pos = hgx2omes2id[hgxc][omesc], hgx2omes2pos[hgxc][omesc]
                if pos != 0 and id_ != 0:
                    csb = (1 - (id_/pos))
                else:
                    csb = 0
                if gcl >= gcl_thresh \
                    and patch >= patch_thresh \
                    and id_ >= id_perc \
                    and pos >= pos_perc \
                    and csb >= csb_thresh:
                    gcfs[hlg] = locs
                    gcf_omes[hlg] = omesc
                    gcf_hgxs[hlg] = hgxc
            except TypeError:
                 if hgx2omes2gcl[hgxc][omesc] >= gcl_thresh \
                    and omes2patch[omesc] >= patch_thresh \
                    and hgx2omes2id[hgxc][omesc] >= id_perc:
                    gcfs[hlg] = locs
                    gcf_omes[hlg] = omesc
                    gcf_hgxs[hlg] = hgxc
            except KeyError:
                eprint(f'\tWARNING: missing from proxies and removed: {omesc, hgxc}')
        return gcfs, gcf_omes, gcf_hgxs
    else:
        return hlgs, hlg_omes, hlg_hgxs

def threshold_hlg_wopos(gcl_thresh, patch_thresh, id_perc, hlgs,
                       hlg_hgxs, hlg_omes, omes2patch,
                       hgx2omes2gcl, hgx2omes2id):
    if any(x > 0 for x in [gcl_thresh, patch_thresh, id_perc]):
        print('\tApplying thresholds', flush = True)
        print('\t\t' + str(len(hlgs)) + ' HLGs before', flush = True)
        gcfs, gcf_omes, gcf_hgxs = {}, {}, {}
        for hlg, locs in hlgs.items():
            hgxc = hlg_hgxs[hlg]
            omesc = hlg_omes[hlg]
            if hgx2omes2gcl[hgxc][omesc] >= gcl_thresh \
                and omes2patch[omesc] >= patch_thresh \
                and hgx2omes2id[hgxc][omesc] >= id_perc:
                gcfs[hlg] = locs
                gcf_omes[hlg] = omesc
                gcf_hgxs[hlg] = hgxc

        return gcfs, gcf_omes, gcf_hgxs
    else:
        return hlgs, hlg_omes, hlg_hgxs


def output_thresholds(logg2d, dist_thresh, gcl_thresh, patch_thresh, id_perc, pos_perc,
                      csb_thresh, hlgs, hlg_hgxs, hlg_omes, omes2dist, omes2patch, hlg2clan,
                      hgx2omes2gcl, hgx2omes2id, hgx2omes2pos, out_dir, i2ome):
    print(f'\tFiltering HLGs for GCFs', flush = True)
    print('\tApplying thresholds', flush = True)
    print('\t\t' + str(len(hlgs)) + ' HLGs before', flush = True)
    hlgs, hlg_omes, hlg_hgxs = threshold_hlg_by_dist(logg2d, dist_thresh, hlgs, hlg_hgxs,
                                                     hlg_omes, omes2dist)
    if hgx2omes2pos:
        gcfs, gcf_omes, gcf_hgxs = threshold_hlg_wpos(gcl_thresh, patch_thresh, id_perc, pos_perc,
                                          csb_thresh, hlgs, hlg_hgxs, hlg_omes, omes2patch,
                                          hgx2omes2gcl, hgx2omes2id, hgx2omes2pos)
        gcf_output = write_hlgs_txt_wpos(gcf_hgxs, gcf_omes, logg2d, 
                            hlg2clan, omes2patch, 
                            hgx2omes2gcl, hgx2omes2id, hgx2omes2pos, i2ome,
                            out_dir, 'gcf')

    else:
        gcfs, gcf_omes, gcf_hgxs = threshold_gcf_wopos(gcl_thresh, patch_thres, 
                                           id_perc, gcfs,
                                           gcf_hgxs, gcf_omes, omes2patch,
                                           hgx2omes2gcl, hgx2omes2id)
        gcf_output = write_hlgs_txt_wopos(gcf_hgxs, gcf_omes, logg2d, 
                            hlg2clan, omes2patch, 
                            hgx2omes2gcl, hgx2omes2id, i2ome,
                            out_dir, 'gcf')


    print(f'\t\t{len(gcfs)} GCFs pass', flush = True)

    return gcf_output, gcfs, gcf_omes, gcf_hgxs


def output_figures(hlg_output, prefix, out_dir):
    # legacy, need to remove unless dn/ds is reimplemented
#    if dnds_dict:
 #       axes = [[],[],[],[],[], []]
  #  else:
    axes = [[],[],[], [], [], []]
    labels = []

    try:
        for entry in hlg_output:
            labels.append(
                ['HLG:', str(entry[1]) + ' | Omes: ' + entry[-1]]
                )
            axes[0].append(entry[5])
            axes[1].append(entry[4])
            axes[2].append(entry[3])
            axes[3].append(entry[6])
            axes[4].append(entry[7])
            axes[5].append(entry[8])
    except ValueError: # no positives
        labels = []
        axes = [[], [], [], []]
        for entry in hlg_output:
            labels.append(
                ['HLG:', str(entry[1]) + ' | Omes: ' + entry[-1]]
                )
            axes[0].append(entry[5])
            axes[1].append(entry[4])
            axes[2].append(entry[3])
            axes[3].append(entry[6])
            
    print(f'\tOutputting {prefix.upper()} plots', flush = True)
    if len(axes) == 4:
        axes_labels = ['PDS', 'GCL', 'nl_TMD',
                       'MMI']
    else:
        axes_labels = ['PDS', 'GCL', 
                       'nl_TMD', 'MMI', 
                       'MMP', 'CSB']

    fig = mk_subplots(labels, axes, axes_labels, alpha = 0.6)
    fig.write_html(out_dir + f'{prefix}_proxies.html')
    fig = mk_3d(labels, axes[:3], axes_labels[:3], alpha = 0.7)
    fig.write_html(out_dir + f'{prefix}s.html')


def quick_thresh(ome, ome_dir, hlgs):
    hlg_f = ome_dir + ome + '/hlg.tsv'
    gcf_f = ome_dir + ome + '/gcf.tsv'
    with open(hlg_f, 'r') as raw, open(gcf_f, 'w') as out:
        for line in raw:
            if line.startswith('#'):
                out.write(line)
            else:
                hlg = int(line.rstrip().split()[3])
                if hlg in hlgs:
                    out.write(line)


def threshold_gcf_quick(db, out_dir, ome_dir, 
                        dist_thresh, gcl_thresh, patch_thresh,
                        id_perc, pos_perc, csb_thresh, cpus = 1):

    ome2hlgs = defaultdict(set)
    print('\tLoading HLGs file', flush = True)
    with gzip.open(out_dir + 'hlgs.tsv.gz', 'rt') as raw, \
        gzip.open(out_dir + 'gcfs.tsv.gz', 'wt') as out:
        for line in raw:
            if line.startswith('#'):
                out.write(line)
            else:
                d = line.rstrip().split()
                hlg = int(d[1])
                tmd = float(d[3])
                pds = float(d[4])
                gcl = float(d[5])
                mmi = float(d[6])
                mmp = float(d[7])
                csb = float(d[8])/100
                omes = d[9].split(',')
                if tmd >= dist_thresh and gcl >= gcl_thresh \
                    and pds >= patch_thresh and mmi >= id_perc \
                    and mmp >= pos_perc and csb >= csb_thresh:
                    for ome in omes:
                        ome2hlgs[ome].add(hlg)
                    out.write(line)

    print('\tExtracting GCFs')    
    with mp.Pool(processes = cpus) as pool:
        pool.starmap(quick_thresh, ((ome, ome_dir, hlgs) \
                                    for ome, hlgs in ome2hlgs.items()))                        


def threshold_gcf_bypass(db, out_dir, wrk_dir, i2ome, gene2hg,
                         dist_thresh, gcl_thresh, patch_thresh,
                         id_perc, pos_perc, csb_thresh, ipr_path,
                         pfam_path, cpus = 1):

    print('\tLoading data structures', flush = True)
    with open(wrk_dir + 'hlgs.pickle', 'rb') as raw:
        hlgs, hlg_omes, hlg_hgxs, hlg2clan = pickle.load(raw)

    with open(wrk_dir + 'pds.full.pickle', 'rb') as in_pick:
        omes2patch = pickle.load(in_pick)

    with open(wrk_dir + 'gcl.pickle', 'rb') as pickin:
        hgx2omes2gcl = pickle.load(pickin)
    with open(wrk_dir + 'mmi.pickle', 'rb') as pickin:
        hgx2omes2id = pickle.load(pickin)
    if os.path.isfile(wrk_dir + 'mmp.pickle'):
        with open(wrk_dir + 'mmp.pickle', 'rb') as pickin:
            hgx2omes2pos = pickle.load(pickin)
    else:
        hgx2omes2pos = None

    with open(wrk_dir + 'omes2dist.pickle', 'rb') as raw:
        omes2dist = pickle.load(raw)

    logg2d = calc_logg2d(hlg_omes, omes2dist)
    gcf_output, gcfs, gcf_omes, gcf_hgxs = output_thresholds(logg2d, dist_thresh, 
                      gcl_thresh, patch_thresh, id_perc, pos_perc, csb_thresh,
                      hlgs, hlg_hgxs, hlg_omes, omes2dist, omes2patch, hlg2clan,
                      hgx2omes2gcl, hgx2omes2id, hgx2omes2pos, out_dir, i2ome)
    output_figures(gcf_output, 'gcf', out_dir)

    
    with mp.Pool(processes = cpus) as pool:
        hlg_gene_res = pool.starmap(read_ome_hlgs, ((ome, f'{out_dir}ome/hlg.tsv') \
                                                    for ome in i2ome))
    hlg_genes = {}
    for ome, ome_hlg_gene in hlg_gene_res:
        if ome_hlg_gene:
            hlg_genes[ome] = ome_hlg_gene

    gcf_genes = write_ome_output('gcf', gcfs, out_dir, db, cpus, gene2hg)
    annotation_mngr(hlg_genes, db, wrk_dir, pfam_path, ipr_path, 
                    gene2hg, out_dir, prefix = 'hlg', cpus = cpus)
    annotation_mngr(gcf_genes, db, wrk_dir, pfam_path, ipr_path, 
                    gene2hg, out_dir, prefix = 'gcf', cpus = cpus)

def write_ome_output(prefix, hlgs, out_dir, db, cpus, gene2hg):
    print(f'\tWriting {prefix.upper()} files', flush = True)
    write_clus_cmds = []
    ome2clusters = defaultdict(list)
    for hlg, loci in hlgs.items():
        for loc in loci:
            ome = loc[0][:loc[0].find('_')]
            ome2clusters[ome].append([loc, hlg])

    ome_dir = out_dir + 'ome/'
    if not os.path.isdir(ome_dir):
        os.mkdir(ome_dir)
    for ome, clusters in ome2clusters.items():
        if not os.path.isdir(ome_dir + ome):
            os.mkdir(ome_dir + ome)
        gff = db[ome]['gff3']
        out_file = ome_dir + ome + f'/{prefix}.tsv'
        write_clus_cmds.append([clusters, ome, out_file, gff, gene2hg])
    with mp.get_context('fork').Pool(processes = cpus) as pool:
        out_genes = pool.starmap(write_clusters, 
                                 tqdm(write_clus_cmds, 
                                      total = len(write_clus_cmds)))
        pool.close()
        pool.join()
    genes = {x[0]: x[1] for x in out_genes if x[1]}
    return genes


def annotation_mngr(hlg_genes, db, wrk_dir, pfam_path, ipr_path, 
                    gene2hg, out_dir, prefix = 'hlg', cpus = 1):
    print('\tSearching', flush = True)
    prot_paths = {}
    for ome in hlg_genes:
        prot_paths[ome] = db[ome]['faa']

    ann_res, failedOmes = ann_mngr(
        hlg_genes, prot_paths, wrk_dir, pfam_path, ipr_path,
        evalue = 0.0001, threshold = 0.5, cpus = cpus
        )
    print('\tAnnotating', flush = True)
    annotate_clusters_cmds = []
    for ome, genes in hlg_genes.items():
        if ome in failedOmes:
            continue
        gff_path = db[ome]['gff3']
        pro_path = db[ome]['faa']
        try:
            annotate_clusters_cmds.append([genes, #gff_path, pro_path, 
                                  ome, out_dir + 'ome/', gene2hg, 
                                  prefix, ann_res[ome]])
        except KeyError:
            pass
#             annotate_clusters_cmds.append([genes, #gff_path, pro_path, 
 #                                  ome, out_dir + 'ome/', gene2hg,
   #                                prefix])


    with mp.get_context('fork').Pool(processes = cpus) as pool:
        pool.starmap(annotate_clusters, 
                             tqdm(annotate_clusters_cmds, 
                                  total = len(annotate_clusters_cmds)))
        pool.close()
        pool.join()


def output_hlgs(db, wrk_dir, hlgs, hlg_omes, i2ome, out_dir, hlg_hgxs,
         omes2dist, omes2patch, hgx2omes2gcl, hgx2omes2id, hgx2omes2pos, 
         gene2hg, plusminus, ome2i, hlg2clan, dist_thresh, gcl_thresh,
         patch_thresh, id_perc, pos_perc, csb_thresh, ipr_path = None,
         pfam_path = None, dnds_dict = {}, cpus = 1):

    print('\tWriting cluster scores', flush = True)

    logg2d = calc_logg2d(hlg_omes, omes2dist)
    if hgx2omes2pos:
        hlg_output = write_hlgs_txt_wpos(hlg_hgxs, hlg_omes, logg2d, 
                            hlg2clan, omes2patch, 
                            hgx2omes2gcl, hgx2omes2id, hgx2omes2pos, i2ome,
                            out_dir, 'hlg')
    else:
        hlg_output = write_hlgs_txt_wopos(hlg_hgxs, hlg_omes, logg2d, 
                            hlg2clan, omes2patch, 
                            hgx2omes2gcl, hgx2omes2id, i2ome,
                            out_dir, 'hlg')

    if dist_thresh or gcl_thresh or patch_thresh or id_perc or pos_perc:
        gcf_output, gcfs, gcf_omes, gcf_hgxs = output_thresholds(logg2d, 
                                   dist_thresh, gcl_thresh, patch_thresh, 
                                   id_perc, pos_perc, csb_thresh, hlgs, hlg_hgxs, 
                                   hlg_omes, omes2dist, omes2patch, hlg2clan,
                                   hgx2omes2gcl, hgx2omes2id, hgx2omes2pos, 
                                   out_dir, i2ome)
    else:
        gcf_output, gcfs, gcf_omes, gcf_hgxs = None, hlgs, hlg_omes, hlg_hgxs

    output_figures(hlg_output, 'hlg', out_dir)
    if gcf_output:
        output_figures(gcf_output, 'gcf', out_dir)

    prefix = 'hlg'
    hlg_genes = write_ome_output('hlg', hlgs, out_dir, db, cpus, gene2hg)
    if gcf_output:
#        prefix = 'gcf'
        gcf_genes = write_ome_output('gcf', gcfs, out_dir, db, cpus, gene2hg)
   
    if pfam_path or ipr_path:    
        print('\nX. Annotating clusters', flush = True)

    annotation_mngr(hlg_genes, db, wrk_dir, pfam_path, ipr_path, 
                    gene2hg, out_dir, prefix = 'hlg', cpus = cpus)
    if dist_thresh or gcl_thresh or patch_thresh or id_perc or pos_perc:
        annotation_mngr(hlg_genes, db, wrk_dir, pfam_path, ipr_path, 
                        gene2hg, out_dir, prefix = 'gcf', cpus = cpus)
